--[[To do:
Managment for special bags
Multiple slot counts
Split sell and delete
Special container learning capability
Close all / expand all
Remove all categories
Recognize disappearance of an item if a new enchanted item appears
List sorted / filtered indicator
Data transfer to other players
Items with charges
Equip Slot indicator
Alternate color indicator for special slot counts
Modify equipped bags
Find solution for "0" item buildup - have to set a lock?
Why didn't I use UISpecialFrames?
If combining stacks will empty a space, check if either is in a special bag
Shift key for forced arrangement
]]
BaudManifestDebug = false;

local PlayerData, DragItem, MailItem, VirtualDrag, BankOpen, TradeOpen, VendorOpen, MailOpen, QueueMail;
local SplitHoldItem, SplitHoldExpire, SplitPending, SplitTimeout, IgnoreHooks, CursorItem;
local CursorHover, RealmData, PlayerName, SelectedChar;
local DisplayNames = {"Inventory","Bank Box"};
--List will refresh on first OnUpdate
local QueuedUpdate = {true, true};
local ItemStringMatch = "(item[%d:]+:(%-?%d+)):(%-?%d+)|";

local FilterDelay = 0.5;
local FadeTime = 0.2;
local IndentSize = 10;
local MailDelay = 0.2;

local CheckButtons = {
  {Text="Show Fading",SavedVar="Fading",Default=true,
    TooltipText="The manifest will fade in when opened, and fade out when closed."},
  {Text="Auto Combine Stacks",SavedVar="Combine",Default=true,
    TooltipText="Automatically combines incomplete item stacks."},
  {Text="Auto Show Inventory",SavedVar="AutoShowInv",Default=true,
    TooltipText="Automatically shows your inventory at the bank, mailbox, vendor, auction house, or while trading."},
  {Text="Stop Blizzard's Auto Show",SavedVar="HideBlizInv",Default=true,
    TooltipText="Stops Blizzard's backpack from automatically opening at the mail and vendor."},
  {Text="Auto Show Bank",SavedVar="ShowBank",Default=true,
    TooltipText="Shows your bank box when it is opened.  The bank box can also be opened with a hotkey, or from the inventory's menu."},
  {Text="Hide Blizzard's Bank",SavedVar="HideBlizBank",Default=true, Depend="ShowBank",
    TooltipText="Prevents Blizzard's bank box from showing when you open your bank box."},
  {Text="Hide Equipped",SavedVar="HideEquipped",Default=true,
    TooltipText="Hides items in the manifest when they are equipped.  Equipped items will always be shown when the list is filtered, sorted, or viewed from another character."},
  {Text="Item Name First",SavedVar="ItemNameFirst",Default=false,
    TooltipText="Displays the item's name before its quantity."}
};

--[[local SliderBars = {
  {Text="Scale - %d%%",Low="50%",High="200%",SavedVar="Scale", Default=100,
    TooltipText = "Scale of the manifest."}
};]]

--UIPanelWindows["BaudManifestCharacters"] = {area="center", pushable = 0, whileDead = 1};
local SaveItemVars = {
  ItemString = true,
  SuffixFactor = true,
  Category = true,
  Count = true,
  Slots = true,
  Equipped = true
};

local Sorts = {
  {Name = "No Sorting"},
  {Name = "Alphabetical", Func = function(ListItem)
    return select(1,GetItemInfo(ListItem.ItemString));
  end},
  {Name = "Rarity", Func = function(ListItem)
    return select(3,GetItemInfo(ListItem.ItemString));
  end},
  {Name = "Item Type", Func = function(ListItem)
    local Name, _, _, _, _, Type, SubType, _, EquipLoc = GetItemInfo(ListItem.ItemString);
    if(EquipLoc=="INVTYPE_ROBE")then
      EquipLoc = "INVTYPE_CHEST";
    elseif(EquipLoc=="INVTYPE_THROWN")then
      EquipLoc = "INVTYPE_RANGED";
    elseif(EquipLoc=="INVTYPE_HOLDABLE")or(EquipLoc=="INVTYPE_SHIELD")then
      EquipLoc = "INVTYPE_WEAPONOFFHAND2";
    end
    return Type..EquipLoc..SubType;
  end},
  {Name = "Slots Used", Func = function(ListItem)
    return -(not ListItem.Count and 0 or ListItem.Slots or 1);
  end}
};

local SpecialBags = {
  ["Soul Bag"] = {6265},
  ["Key Ring"] = {"Key"},  --Key Ring is a name I assigned
  ["Quiver"] = {"Arrow"},
  ["Ammo Pouch"] = {"Bullet"},
  ["Toolbox"] = {"Parts","Devices","Explosives"}
};


local function DebugMsg(Msg)
  if BaudManifestDebug then
    DEFAULT_CHAT_FRAME:AddMessage(Msg);
  end
end


StaticPopupDialogs["BAUDMANIFEST_DELETECATEGORY"] = {
  text = "Are you sure you want to delete this category?",
  button1 = TEXT(YES),
  button2 = TEXT(NO),
  OnAccept = function()
    BaudManifestDeleteDraggedCategory();
  end,
  OnCancel = function()
    DebugMsg("Cancel delete category.");
    BaudManifestClearVirtualDrag();
  end,
  timeout = 0,
  exclusive = 1,
  hideOnEscape = 1
};


StaticPopupDialogs["BAUDMANIFEST_CATEGORIZE"] = {
  text = "This action will put uncategorized items into categories based on their type.  Do you wish to proceed?",
  button1 = TEXT(YES),
  button2 = TEXT(NO),
  OnAccept = function(Display)
    BaudManifestCategorize(Display);
  end,
  timeout = 0,
  exclusive = 1,
  hideOnEscape = 1
};


local function IsDisplayAccessable(Display)
  return (Display:GetID()==1)or BankOpen and(Display:GetID()==2);
end


function BaudManifestToggleDisplay(Display)
  Display = getglobal("BaudManifestDisplay"..Display);
  if Display:IsShown()then
    Display:Hide();
  else
    Display:Show();
  end
end


function BaudManifestToggleChars()
  if BaudManifestCharacters:IsShown()then
    BaudManifestCharacters:Hide()
  else
    BaudManifestCharacters:Show();
  end
end


function BaudManifest_OnLoad()
  BaudManifestVersionText:SetText("Version "..GetAddOnMetadata("BaudManifest","Version"));

  SlashCmdList["BaudManifest"] = function(Msg)
    if(strlower(Msg)=="options")then
      BaudManifestOptions:Show();
    else
      BaudManifestToggleDisplay(1);
    end
  end
  SLASH_BaudManifest1 = "/baudmanifest";
  SLASH_BaudManifest2 = "/baudman";
  SLASH_BaudManifest3 = "/bm";
  SLASH_BaudManifestMenuName = "Baud Manifest";
  SLASH_BaudManifestMenuArgs = "Options";
  BINDING_HEADER_BaudManifest = "Baud Manifest";
  BINDING_NAME_BaudManifest = "Toggle Inventory";
  BINDING_NAME_BaudManifestToggleBank = "Toggle Bank";
  BINDING_NAME_BaudManifestToggleChars = "Toggle Characters";
  this:RegisterEvent("VARIABLES_LOADED");
  this:RegisterEvent("BAG_UPDATE");
  this:RegisterEvent("UNIT_INVENTORY_CHANGED");
  this:RegisterEvent("BAG_CLOSED");
  this:RegisterEvent("BAG_UPDATE_COOLDOWN");
  this:RegisterEvent("PLAYERBANKSLOTS_CHANGED");
  this:RegisterEvent("BANKFRAME_OPENED");
  this:RegisterEvent("BANKFRAME_CLOSED");
  this:RegisterEvent("TRADE_SHOW");
  this:RegisterEvent("TRADE_CLOSED");
  this:RegisterEvent("TRADE_PLAYER_ITEM_CHANGED");
  this:RegisterEvent("MAIL_SHOW");
  this:RegisterEvent("MAIL_CLOSED");
  this:RegisterEvent("MAIL_SEND_SUCCESS");
  this:RegisterEvent("MAIL_FAILED");
  this:RegisterEvent("MERCHANT_SHOW");
  this:RegisterEvent("MERCHANT_CLOSED");
  this:RegisterEvent("AUCTION_HOUSE_SHOW");
  this:RegisterEvent("AUCTION_HOUSE_CLOSED");
  this:RegisterEvent("CURSOR_UPDATE");
  this:RegisterEvent("ITEM_LOCK_CHANGED");
  this:RegisterEvent("PLAYER_MONEY");
  this:RegisterEvent("PLAYER_LOGIN");
  this:RegisterEvent("PLAYER_LOGOUT");
  DEFAULT_CHAT_FRAME:AddMessage("Baud Manifest: AddOn loaded. Type /baudman to display your manifest.");
end


local function GetCharData(Name)
  if(type(RealmData[Name])~="table")then
    RealmData[Name] = {};
  end
  local CharData = RealmData[Name];
  for Index = 1, 2 do
    if(type(CharData[Index])~="table")then
      CharData[Index] = {};
    end

    if(type(CharData[Index][1])~="table")then
      CharData[Index][1] = {};
    end

    if(type(CharData[Index].BGColor)~="table")then
      CharData[Index].BGColor = {0.5,0.5,0.5,0.7};
    end
  end
  return CharData;
end


function BaudManifest_OnEvent()
  if(event=="VARIABLES_LOADED")then
    if(type(BaudManifestData)~="table")then
      BaudManifestData = {};
    end
    local RealmName = GetRealmName();
    if(type(BaudManifestData[RealmName])~="table")then
      BaudManifestData[RealmName] = {};
    end
    RealmData = BaudManifestData[RealmName];
    PlayerName = UnitName("player");
    --Copy from individual character saved variables for backwards compatibility
    if not RealmData[PlayerName]and BaudManifest_Cfg and BaudManifest_Cfg.Items then
      RealmData[PlayerName] = BaudManifest_Cfg.Items;
    end
    PlayerData = GetCharData(PlayerName);
    if(type(PlayerData.Scale)~="number")then
      PlayerData.Scale = 100;
    end
    BaudManifestUpdateScales();

    for Index = 1, 2 do
      local Display = getglobal("BaudManifestDisplay"..Index);
      Display.Data = PlayerData[Index];
      BaudManifestUpdateBGColor(Display);
    end
    BaudManifestUpdateBlizBankBlock();
    local CharList = {};
    for Name, Value in pairs(RealmData)do
      if(Name~=PlayerName)and(Name~="All Characters")then
        tinsert(CharList,Name);
      end
    end
    table.sort(CharList,function(a,b) return(a < b);end);
    tinsert(CharList,1,"All Characters");
    local Entry, Texture, Coords, Text;
    for Key, Value in ipairs(CharList)do
      Entry = CreateFrame("Button","BaudManifestCharacterEntry"..Key,BaudManifestCharactersScrollBox,"BaudManifestCharacterEntryTemplate");
      Entry:SetPoint("TOP",0,6-16*Key);
      Text = getglobal(Entry:GetName().."Text");
      Text:SetText(Value);
      Entry.Name = Value;
      if(Key==1)then
        Text:SetTextColor(1,1,1);
        Text:SetPoint("LEFT",2,0);
      else
        Texture = getglobal(Entry:GetName().."Texture");
        Coords = CLASS_BUTTONS[strupper(RealmData[Value].Class or "")];
        if Coords then
      	  Texture:SetTexture("Interface\\WorldStateFrame\\Icons-Classes");
    	    Texture:SetTexCoord(unpack(Coords));
    	    Texture:Show();
        end
      end
    end

    if(type(PlayerData.PutQuality)~="table")then
      PlayerData.PutQuality = {};
    end
    for Quality = 0, 2 do
      if not PlayerData.PutQuality[Quality]then
        PlayerData.PutQuality[Quality] = 1;
      end
    end
    for Index, Value in ipairs(CheckButtons)do
      if(PlayerData[Value.SavedVar]==nil)then
        PlayerData[Value.SavedVar] = Value.Default;
      end
    end

  elseif(event=="PLAYER_LOGIN")then
    PlayerData.Class = UnitClass("player");
    PlayerData.Money = GetMoney();

  elseif(event=="PLAYER_LOGOUT")then
    --Variable cleanup here
    for Display, Value in ipairs(PlayerData)do
      for Category, Value in ipairs(Value)do
        for ItemIndex, Value in ipairs(Value)do
          for Key, _ in pairs(Value)do
            if not SaveItemVars[Key]then
              Value[Key] = nil;
            end
          end
        end
      end
    end

  elseif(event=="PLAYER_MONEY")then
    PlayerData.Money = GetMoney();

  elseif(event=="BAG_UPDATE")or(event=="BAG_CLOSED")or(event=="PLAYERBANKSLOTS_CHANGED")then
    --Updating of items done even if the bag is closed to keep the ordering correct
    local Display;
    if(event=="PLAYERBANKSLOTS_CHANGED")or(arg1==-1)or(arg1 > 4)then
      Display = 2;
    else
      Display = 1;
    end
    QueuedUpdate[Display] = true;
    if DragItem and DragItem.Pending and(DragItem.Origin:GetID()==Display)then
      DragItem.Pending = 1;
    end

  elseif(event=="UNIT_INVENTORY_CHANGED")then
    if(arg1~="player")then
      return;
    end
    QueuedUpdate[1] = true;

  elseif(event=="ITEM_LOCK_CHANGED")then
    --The list is updated to change the specific targeted items, but only if the list is open
    for Display = 1, 2 do
      if getglobal("BaudManifestDisplay"..Display):IsShown()then
        QueuedUpdate[Display] = true;
      end
    end

  elseif(event=="BAG_UPDATE_COOLDOWN")then
    BaudManifestScrollBar_Update(BaudManifestDisplay1);

  elseif(event=="BANKFRAME_OPENED")then
    BankOpen = true;
    BaudManifestUpdateList(BaudManifestDisplay2);
    BaudManifestAutoShow(false, true);

  elseif(event=="BANKFRAME_CLOSED")then
    BankOpen = false;
    BaudManifestAutoShow(true, true);

  elseif(event=="TRADE_SHOW")then
    TradeOpen = true;
    BaudManifestPickupDragRemain(BaudManifestTradeRemain);
    BaudManifestAutoShow();

  elseif(event=="TRADE_CLOSED")then
    TradeOpen = false;
    BaudManifestAutoShow(true);

  elseif(event=="TRADE_PLAYER_ITEM_CHANGED")then
    if DragItem and not DragItem.MoveMethod then
      BaudManifestPickupDragRemain(BaudManifestTradeRemain);
    end

  elseif(event=="MAIL_SHOW")then
    MailOpen = true;
    BaudManifestAutoShow();

  elseif(event=="MAIL_CLOSED")then
    MailOpen = false;
    MailItem = nil;
    BaudManifestAutoShow(true);

  elseif(event=="MAIL_SEND_SUCCESS")then
    DebugMsg("Send Success");
    --The successive mail sends will only work if done later than this, and most reliably with a delay
    QueueMail = GetTime() + MailDelay;

  elseif(event=="MAIL_FAILED")then
    DebugMsg("Send Failed");

  elseif(event=="MERCHANT_SHOW")then
    VendorOpen = true;
    BaudManifestAutoShow();

  elseif(event=="MERCHANT_CLOSED")then
    VendorOpen = false;
    BaudManifestAutoShow(true);

  elseif(event=="AUCTION_HOUSE_SHOW")then
    BaudManifestAutoShow();

  elseif(event=="AUCTION_HOUSE_CLOSED")then
    BaudManifestAutoShow(true);

  elseif(event=="CURSOR_UPDATE")then
    if(arg1=="up")then
      BaudManifestClearFocus();
    end
    if(CursorHasItem()~=CursorItem)then
      CursorItem = CursorHasItem();
      if not CursorItem then
        BaudManifestDragCount:Hide();
--[[        if DragItem and not DragItem.Remain then
          DragItem = nil;
        end]]
      elseif DragItem and not IgnoreHooks then
        DragItem = nil;
      end
      BaudManifestUpdateBlocked();
    end
  end
end


function BaudManifestAutoShow(Hide, Bank)
  local Display;
  for Index = 1, 2 do
    if(Index==1)and PlayerData.AutoShowInv or(Index==2)and Bank then
      Display = getglobal("BaudManifestDisplay"..Index);
      if Hide then
        if Display:IsShown()then
          if Display.AutoShown then
            Display:Hide();
          else
            BaudManifestScrollBar_Update(Display);
          end
        end
      elseif((Index~=2)or(PlayerData.ShowBank~=false))and(not Display:IsShown()or Display.Closing)then
        Display:Show();
        Display.Closing = false;
        Display.AutoShown = true;
      end
    end
  end
end


function BaudManifestCategoryEditBox_OnEditFocusLost()
  if this:IsShown()then
    this.Category.Name = this:GetText();
  end
  this:Hide();
  BaudManifestScrollBar_Update(this.Display);
end


function BaudManifestClearVirtualDrag()
  if not VirtualDrag then
    return;
  end
  DebugMsg("Clearing virtual drag.");
  PlaySound("INTERFACESOUND_CURSORDROPOBJECT");
  VirtualDrag = nil;
  BaudManifestVirtualDrag:Hide();
  BaudManifestUpdateBlocked();
end


function BaudManifestDeleteDraggedCategory()
  if not VirtualDrag or not VirtualDrag.Category then
    DebugMsg("No category to delete.");
    return;
  end
  DebugMsg("Deleting category #"..VirtualDrag.Category);

  local Replace, ListIndex;
  for Category, ParentCat in ipairs(VirtualDrag.Origin.Data)do
    ListIndex = 1;
    while ParentCat[ListIndex]do
      if ParentCat[ListIndex].Category then
        if(ParentCat[ListIndex].Category==VirtualDrag.Category)then
          DebugMsg("Located category listing, removing and replacing it with its contents.");
          tremove(ParentCat, ListIndex);
          ListIndex = ListIndex - 1;
          --Now replace the deleted category with its contents
          if not Replace then
            Replace = true;
            for Key, Value in ipairs(VirtualDrag.Origin.Data[VirtualDrag.Category])do
              --Make sure the copies entries are skipped, or categories could be decremented twice
              ListIndex = ListIndex + 1;
              tinsert(ParentCat,ListIndex,Value);
            end
          end
        elseif(ParentCat[ListIndex].Category > VirtualDrag.Category)then
          ParentCat[ListIndex].Category = ParentCat[ListIndex].Category - 1;
        end
      end
      ListIndex = ListIndex + 1;
    end
  end
  if(VirtualDrag.Origin:GetID()==1)then
    for Quality, Value in pairs(PlayerData.PutQuality)do
      if(Value==VirtualDrag.Category)then
        PlayerData.PutQuality[Quality] = 1;
      elseif(Value > VirtualDrag.Category)then
        PlayerData.PutQuality[Quality] = PlayerData.PutQuality[Quality] - 1;
      end
    end
  end

  DebugMsg("Deleting category variable.");
  tremove(VirtualDrag.Origin.Data,VirtualDrag.Category);
  BaudManifestUpdateDisplayList(VirtualDrag.Origin);
  BaudManifestClearVirtualDrag();
end


function BaudManifestCategoriesDisabled(Display)
  return (Display.Sort~=1)or(Display.Filter~=1)or strfind(Display.CustomFilter or "","[^%s]");
end


function BaudManifestCategorize(Display)
  DebugMsg("Categorizing loose items");
  local Type, SubType, Category, New;
  local Index = 1;
  while(Display.Data[1][Index])do
    if Display.Data[1][Index].ItemString then
      _, _, _, _, _, Type, SubType = GetItemInfo(Display.Data[1][Index].ItemString);
      if not Type then
        Index = Index + 1;
      else
        Category, New = BaudManifestGetCategory(Display, Type, 1);
        if New then
          Index = Index + 1;
        end
        tinsert(Display.Data[Category],Display.Data[1][Index]);
        tremove(Display.Data[1],Index);
      end
    else
      Index = Index + 1;
    end
  end
  BaudManifestUpdateDisplayList(Display);
end


--First argument is the only one required
function BaudManifestGetCategory(Display, Name, ParentCategory)
  if BaudManifestCategoriesDisabled(Display)then
    UIErrorsFrame:AddMessage("Cannot create categories while the list is sorted or filtered.", 1.0, 0.1, 0.1, 1.0, UIERRORS_HOLD_TIME);
    return;
  end
  if ParentCategory then
    for Key, Value in ipairs(Display.Data[ParentCategory])do
      if Value.Category and(Display.Data[Value.Category].Name==Name)then
        return Value.Category;
      end
    end
  end
  if not Name then
    Name = "New Category";
  end
  DebugMsg("Creating category: "..Name);
  tinsert(Display.Data,{Name=Name, Expanded=false});
  local Category = #Display.Data;
  tinsert(Display.Data[(ParentCategory or 1)],1,{Category = Category});
  FauxScrollFrame_SetOffset(getglobal(Display:GetName().."ScrollBar"),0);
  if not ParentCategory then
    BaudManifestUpdateDisplayList(Display);
  end
  return Category, true;
end


function BaudManifestUpdateBlocked()
  --Enabling and disabling item entries to serve this purpose will prevent items from being dropped by releasing the cursor
  local Enable = CursorItem or VirtualDrag;
  for Display = 1, 4 do
    local Frame = getglobal("BaudManifestDisplay"..Display.."ClickBlock");
    if Enable then
      Frame:Show();
    else
      Frame:Hide();
    end
  end
  DebugMsg("Click block "..(Enable and "Enabled." or "Disabled."));
end


hooksecurefunc("PickupContainerItem",function(Bag,Slot)
  if IgnoreHooks or DragItem or not CursorHasItem()then
    return;
  end
  DragItem = GetContainerItemLink(Bag, Slot);
  if(DragItem~=select(3,GetCursorInfo()))then
    DragItem = nil;
    return;
  end
  DragItem = string.match(DragItem, ItemStringMatch);
  if not DragItem then
    return;
  end
  DragItem = {String = DragItem, Stack = select(2,GetContainerItemInfo(Bag,Slot)), Origin = getglobal("BaudManifestDisplay"..(((Bag == -1)or(Bag > 4))and 2 or 1))};
  BaudManifestShowDragCount();
  DebugMsg("Item pickup from "..Bag..", "..Slot);
end);


--Used for deleted multiple items at once.  Use caution!
hooksecurefunc("DeleteCursorItem",function()
  if not IgnoreHooks then
    BaudManifestPickupDragRemain(function()
      --This check is used as a failsafe
      local Type, _, Link = GetCursorInfo();
      if(Type~="item")then
        return;
      end
      if(DragItem.String~=string.match(Link,ItemStringMatch))then
        DEFAULT_CHAT_FRAME:AddMessage("Baud Manifest: Item match failure.",1,0.1,0.1);
        return;
      end
      IgnoreHooks = true;
      DeleteCursorItem();
      IgnoreHooks = false;
      return true;
    end, true);
  end
end);


--This function is used when dragging items onto the vendor window.  The values passed to it don't matter in this case.
hooksecurefunc("PickupMerchantItem",function()
  if not IgnoreHooks and VendorOpen then
    BaudManifestPickupDragRemain(function()
      if VendorOpen then
        IgnoreHooks = true;
        PickupMerchantItem();
        IgnoreHooks = false;
        return true;
      end
      return false;
    end, true);
  end
end);


function BaudManifestTradeRemain()
  if TradeOpen then
    IgnoreHooks = true;
    DropItemOnUnit("NPC");
    IgnoreHooks = false;
    return true;
  end
  return false;
end


hooksecurefunc("ClickSendMailItemButton",function()
  if IgnoreHooks or not MailOpen then
    return;
  end
  --Remember that clicking doesn't mean the item will be accepted (soulbound / quest)
  if DragItem and(GetSendMailItem()~=GetItemInfo(DragItem.String..":"..(DragItem.SuffixFactor or 0)))then
    DebugMsg("Mail item mismatch.");
    return;
  end
  if MailItem and(not CursorHasItem()or(MailItem.String~=string.match(select(3, GetCursorInfo())or"",ItemStringMatch)))then
    MailItem = nil;
  end

  DragItem, MailItem = MailItem, DragItem;
  if DragItem then
    BaudManifestShowDragCount();
  end
  if not MailItem then
    SendMailFrame_Update();
  else
    BaudManifestUpdateMailItemCount();
  end
end);


function BaudManifestUpdateMailItemCount()
  if not MailItem or not MailItem.Remain or not GetSendMailItem()then
    return;
  end
  SendMailPackageButtonCount:SetText(MailItem.Stack + MailItem.Remain);
end


hooksecurefunc("SendMailFrame_Update",function()
  BaudManifestUpdateMailItemCount();
end);


hooksecurefunc("SendMail", function(...)
  local SendData = {...};
  DebugMsg("Send Mail: "..SendData[1]..", "..SendData[2]);
  if IgnoreHooks or not MailOpen or not MailItem or not MailItem.Remain then
    return;
  end
  MailItem.SendData = SendData;
  MailItem.ChangeSubject = not SendData[2]or strfind(SendData[2],"^"..(GetItemInfo(MailItem.String)or ""));
  MailItem.MoveMethod = function()
    DragItem = nil;
    if not MailItem then
      return;
    end
    IgnoreHooks = true;
    ClickSendMailItemButton();
    IgnoreHooks = false;
    if MailItem.ChangeSubject then
      local Name, _, Count = GetSendMailItem();
      MailItem.SendData[2] = Name;
      if(Count > 1)then
        SendData[2] = SendData[2].." ("..Count..")";
      end
    end
    IgnoreHooks = true;
    SendMail(unpack(MailItem.SendData));
    IgnoreHooks = false;
    return false;
  end
end);


hooksecurefunc("DropItemOnUnit",function(Unit)
  if not IgnoreHooks and TradeOpen and UnitIsUnit("NPC",Unit)then
    BaudManifestPickupDragRemain(BaudManifestTradeRemain);
  end
end);


hooksecurefunc("ClickTradeButton",function(Index)
  if TradeOpen then
    BaudManifestPickupDragRemain(BaudManifestTradeRemain);
  end
end);


function BaudManifestDragCount_OnUpdate()
  local X, Y = GetCursorPosition();
  local Scale = this:GetEffectiveScale();
  this:ClearAllPoints();
  this:SetPoint("TOPLEFT",UIParent,"BOTTOMLEFT", X /Scale - 5, Y /Scale - 27);
end


function BaudManifestVirtualDrag_OnUpdate()
  --Dragged categories will not be cleared for stuff that is dragged and immediately placed, but if anything is in the cursor
  --for an update, then the category is dropped.
  if GetCursorInfo()then
    BaudManifestClearVirtualDrag();
    return;
  end
  this:ClearAllPoints();
  this:SetPoint("BOTTOMRIGHT",UIParent,"BOTTOMLEFT",GetCursorPosition());
end


function BaudManifestVirtualDragClickBlock_OnClick()
  if(arg1=="RightButton")or not VirtualDrag.Category then
    BaudManifestClearVirtualDrag();
  else
    StaticPopup_Show("BAUDMANIFEST_DELETECATEGORY");
  end
end


function BaudManifestShowDragCount()
  local Count = DragItem.Stack;
  if DragItem.Remain then
    Count = Count + DragItem.Remain;
  end
  if not CursorHasItem()or(Count==1)and(select(8,GetItemInfo(DragItem.String))==1)then
    return;
  end
  BaudManifestDragCountText:SetText(Count);
  DebugMsg("Showing drag count of "..Count);
  BaudManifestDragCount:Show();
end


function BaudManifestBagPriority(Bag,Item)
  --nil - Item cannot go in this bag, 0 - Unknown, 1 - Any item could go here, 2 - This bag is specificly for this item type
  local BagType;
  if(Bag == -2)then
    BagType = "Key Ring";
  else
    if(Bag < 1)then
      return 1;
    end
    BagType = GetInventoryItemLink("player",ContainerIDToInventoryID(Bag));
    if not BagType then
      return;
    end
    BagType = select(7, GetItemInfo(BagType));
    if not BagType then
      return;
    end
  end
  if(BagType=="Bag")then
    return 1;
  elseif not SpecialBags[BagType]or not Item then
    return 0;
  end
  local ItemType = select(7, GetItemInfo(Item));
  if not ItemType then
    return;
  end
  for Key, Value in ipairs(SpecialBags[BagType])do
    if(Value==ItemType)then
      return 2;
    end
  end
  return;
end


function BaudManifest_OnUpdate()
  if DragItem and not DragItem.Pending and not CursorItem then
    DebugMsg("Clearing dragged item.");
    DragItem = nil;
    BaudManifestDropLine:Hide();
  end
  if CursorHover and(DragItem and(CursorHover==DragItem.Origin)or VirtualDrag and(CursorHover==VirtualDrag.Origin))then
    local Position, Indent, _, Index = BaudManifestGetCursorPos(CursorHover);
    local ScrollBar = getglobal(CursorHover:GetName().."ScrollBar");
    BaudManifestDropLine:SetParent(CursorHover);
    BaudManifestDropLine:ClearAllPoints();
    local Y = 16 * (1 - Position);
    BaudManifestDropLine:SetPoint("TOP",ScrollBar,"TOP", 0, Y + 1);
    BaudManifestDropLine:SetPoint("LEFT",ScrollBar,"LEFT",Indent * IndentSize,0);
    BaudManifestDropLine:SetPoint("RIGHT",ScrollBar);
    if not Index then
      BaudManifestDropLine:SetPoint("BOTTOM",ScrollBar,"TOP", 0, Y - 16);
    end
    BaudManifestDropLine:SetFrameLevel(ScrollBar:GetFrameLevel() + 3);
    BaudManifestDropLine:Show();
  end
  if UpdateSizes then
    UpdateSizes = false;
    for Display = 1, 4 do
      BaudManifestDisplay_OnSizeChanged(getglobal("BaudManifestDisplay"..Display));
    end
  end

  if QueueMail and(GetTime() > QueueMail)then
    QueueMail = nil;
    if not MailItem or not MailItem.SendData or not MailOpen or GetSendMailItem()then
      return;
    end
    if not MailItem.Remain then
      MailItem = nil;
      return;
    end
    ClearCursor();
    DragItem = MailItem;
    BaudManifestPickupDragRemain();
  end

  for Display = 1, 2 do
    if QueuedUpdate[Display]then
      QueuedUpdate[Display] = false;
      if DragItem and(DragItem.Pending==1)and(DragItem.Origin:GetID()==Display)then
        DragItem.Pending = nil;
        --if the split item is picked up immediately, the client will not register that it is locked
        BaudManifestPickupDragRemain();
      end
      BaudManifestUpdateList(getglobal("BaudManifestDisplay"..Display));
    end
  end
end


function BaudManifestForEachSlot(Display, Func)
  if(Display:GetID()==1)then
    for Bag = 0, 4 do
      for Slot = 1, GetContainerNumSlots(Bag)do
        if Func(Bag, Slot)then
          return true;
        end
      end
    end
    --Equipped items
    for Slot = 1, 19 do
      if Func(nil, Slot)then
        return true;
      end
    end
    --Bliz has a function to say if you have a key ring, so hey, why not?
    if HasKey()then
      for Slot = 1, GetContainerNumSlots(-2)do
        if Func(-2, Slot)then
          return true;
        end
      end
    end
  else
    for Slot = 1, NUM_BANKGENERIC_SLOTS do
      if Func(-1, Slot)then
        return true;
      end
    end
    for Bag = 5, 4 + GetNumBankSlots() do
      for Slot = 1, GetContainerNumSlots(Bag)do
        if Func(Bag, Slot)then
          return true;
        end
      end
    end
  end
  return false;
end


local function AnchorDropdown(DropDown)
  getglobal(DropDown:GetName().."Middle"):SetPoint("RIGHT",-25,0);
  getglobal(DropDown:GetName().."Text"):SetPoint("LEFT",DropDown:GetName().."Left","RIGHT",7,0);
  DropDown.noResize = 1;
end


--This method is used to ensure that esc will close the color picker before the manifest
local pre_CloseMenus = CloseMenus;
CloseMenus = function()
  local Closed;
  Closed = pre_CloseMenus();
  if Closed then
    return Closed;
  end
  local Display;
  for Index = 1, 2 do
    Display = getglobal("BaudManifestDisplay"..Index);
    if Display:IsShown()and not Display.Closing then
      Display:Hide();
      Closed = true;
    end
  end
  return Closed;
end


function BaudManifestDisplay_OnLoad()
  this:RegisterForDrag("LeftButton");
  this:RegisterForClicks("LeftButtonUp","RightButtonUp");

  if(this:GetID()%2==1)then
    --Remove the gap on displays with no money frame on the bottom
    getglobal(this:GetName().."FilterEditBox"):SetPoint("BOTTOM",0,26);
  end
--  tinsert(UISpecialFrames,this:GetName());
  this.List = {};
  this.Sort = 1;
  this.Filter = 1;
  this.CustomFilter = nil;

  local DropDown = getglobal(this:GetName().."SortDropDown");
  AnchorDropdown(DropDown);
  UIDropDownMenu_Initialize(DropDown, BaudManifestSortDropDown_Initialize);
  UIDropDownMenu_SetSelectedValue(DropDown, this.Sort);

  local DropDown = getglobal(this:GetName().."FilterDropDown");
  AnchorDropdown(DropDown);
  UIDropDownMenu_Initialize(DropDown, BaudManifestFilterDropDown_Initialize);
  UIDropDownMenu_SetSelectedValue(DropDown, this.Filter);

  getglobal(this:GetName().."ClickBlock"):SetFrameLevel(this:GetFrameLevel() + 4);

  --Can't be done in XML because the text must be made before the anchor frame is
  getglobal(this:GetName().."NoEntries"):SetPoint("CENTER",this:GetName().."ScrollBar");
  getglobal(this:GetName().."Text"):SetPoint("RIGHT",this:GetName().."OptionsButton","LEFT");

  local Divider = getglobal(this:GetName().."Divider");
  Divider:SetTexCoord(0,1,0.125,1,0,0,0.125,0);
  Divider:SetVertexColor(0.5,0.5,0.5);
end


function BaudManifestOptionsDropDown_OnLoad()
  UIDropDownMenu_Initialize(this, BaudManifestOptionsDropDown_Initialize, "MENU");
end


function BaudManifestDisplay_OnUpdate()
  if this.FadeStart then
    local Alpha = (GetTime() - this.FadeStart) / FadeTime;
    if this.Closing then
      Alpha = 1 - Alpha;
      if(Alpha < 0)then
        this.FadeStart = nil;
        this:Hide();
        this.Closing = nil;
        return;
      end
    elseif(Alpha > 1)then
      this:SetAlpha(1);
      this.FadeStart = nil;
      return;
    end
    this:SetAlpha(Alpha);
  end
  if this.Refresh and(this.Refresh < GetTime())then
    DebugMsg("Refreshing "..this:GetName());
    if this.UpdateList then
      BaudManifestUpdateDisplayList(this);
    else
      BaudManifestScrollBar_Update(this);
    end
  end
  if this.ClearZero then
    local Scale = this:GetEffectiveScale();
    local X, Y = GetCursorPosition();
    X, Y = X / Scale, Y / Scale;
    if(X < this:GetLeft())or(X > this:GetRight())or(Y < this:GetBottom()or(Y > this:GetTop()))then
      BaudManifestClearZeros(this);
      BaudManifestUpdateDisplayList(this);
    end
  end
  --Fix for Blizzard bug
  if this.UpdateHeight then
    this.UpdateHeight = false;
    BaudManifestDisplay_OnSizeChanged(this);
  end
end


function BaudManifestClearZeros(Display)
  Display.ClearZero = nil;
  local Index, Delete;
  for Category, Value in ipairs(Display.Data)do
    Index = 1;
    while(Value[Index])do
      if Value[Index].ItemString and not Value[Index].Count then
        if(Category==1)or(Display:GetID()==1)and(PlayerData.PutQuality[select(3,GetItemInfo(Value[Index].ItemString))or -1]==Category)then
          tremove(Value,Index);
          Index = Index - 1;
        elseif Value[Index].Show then
          Value[Index].Show = nil;
        end
      end
      Index = Index + 1;
    end
  end
end


function BaudManifestDisplay_OnShow()
  if this.Shown then
    return;
  end
  this.Shown = true;
  if this.FadeStart then
    return;
  end
  if(PlayerData.Fading~=false)then
    this.FadeStart = GetTime();
  else
    this:SetAlpha(1);
  end
  this.Closing = false;
  PlaySound("igBackPackOpen");
  local Level = this:GetFrameLevel() + 1;
  local Entry;
  for Index = 1, this.Entries do
    Entry = getglobal(this:GetName().."Entry"..Index);
    Entry:SetFrameLevel(Level);
    getglobal(Entry:GetName().."BlizItem"):SetFrameLevel(Level + 1);
    getglobal(Entry:GetName().."BlizInv"):SetFrameLevel(Level + 1);
  end
  BaudManifestClearZeros(this);
  BaudManifestUpdateDisplayList(this);
end


function BaudManifestDisplay_OnHide()
  if this:IsShown()then
    return;
  end
  this.Shown = nil;
  if this.Closing then
    if this.FadeStart then
      this:Show();
    end
    return;
  end
  if(PlayerData.Fading~=false)then
    this.FadeStart = GetTime();
    this.Closing = true;
    this:Show();
  end
  PlaySound("igBackPackClose");
  if(this:GetID()==2)and BankOpen and this.AutoShown then
    CloseBankFrame();
  end
  this.AutoShown = nil;
end


function BaudManifestDisplay_OnSizeChanged(Display)
  local ScrollBar = getglobal(Display:GetName().."ScrollBar");
  local Entries = floor(ScrollBar:GetHeight()/16);
  if(Entries == Display.Entries)then
    return;
  end
  Display.Entries = Entries;
  local Button;
  if not Display.MaxEntries or(Display.MaxEntries < Entries)then
    for Index = (Display.MaxEntries or 0) + 1, Entries do
      Button = CreateFrame("Button",Display:GetName().."Entry"..Index,Display,"BaudManifestItemEntryTemplate");
      --Showing the texture affects the text positioning for some reason, and cannot be done in OnLoad
      getglobal(Button:GetName().."Texture"):Show();
      Button:ClearAllPoints();
      Button:SetPoint("TOP",ScrollBar,"TOP",0,16*(1-Index));
      Button:SetPoint("LEFT",ScrollBar);
      Button:SetPoint("RIGHT",ScrollBar);
      Button.Display = Display;
    end
    Display.MaxEntries = Entries;
  else
    for Index = (Entries + 1), Display.MaxEntries do
      getglobal(Display:GetName().."Entry"..Index):Hide();
    end
  end
  BaudManifestScrollBar_Update(Display)
end


function BaudManifestClearFocus()
  BaudManifestCategoryEditBox:ClearFocus();
  for Display = 1, 4 do
    getglobal("BaudManifestDisplay"..Display.."FilterEditBox"):ClearFocus();
  end
end


function BaudManifestDisplay_OnClick(Display)
  DebugMsg("Click on "..Display:GetName());
  BaudManifestClearFocus();
  if(arg1=="RightButton")then
    BaudManifestClearVirtualDrag();
  else
    BaudManifestDropItem(Display);
  end
end


function BaudManifestDisplay_OnDragStart(Display)
  if Display:IsMovable()then
    Display:StartMoving();
  end
end


function BaudManifestDisplay_OnDragStop()
  this:StopMovingOrSizing();
end


function BaudManifestDisplay_OnEnter(Display)
  CursorHover = Display;
end


function BaudManifestDisplay_OnLeave()
  BaudManifestDropLine:Hide();
  CursorHover = nil;
end


function BaudManifestResize()
  local Display = this:GetParent();
  if Display.Data.LockSize or not Display:IsResizable()then
    BaudManifestDisplay_OnDragStart(Display);
    return;
  end
  local AnchorPoint = strupper(strsub(this:GetName(),strlen(Display:GetName().."Resize")+1));
  Display:StartSizing(AnchorPoint);
end


function BaudManifestStopResize()
  this:GetParent():StopMovingOrSizing();
end


function BaudManifestItemEntry_OnLoad()
  this:RegisterForClicks("LeftButtonUp","RightButtonUp");
  this:RegisterForDrag("LeftButton");
  this:RegisterEvent("BAG_UPDATE_COOLDOWN");
  this:RegisterEvent("MODIFIER_STATE_CHANGED");
end


function BaudManifestItemEntry_OnEvent()
  if GameTooltip:IsOwned(this)then
    BaudManifestItemEntry_OnEnter(this);
  end
end


function BaudManifestItemEntry_OnUpdate(Button)
--[[  if GameTooltip:IsOwned(Button) and Button.updateTooltip then
    Button.updateTooltip = Button.updateTooltip - arg1;
    if(Button.updateTooltip > 0 )then
      return;
    end
    BaudManifestItemEntry_OnEnter(Button, true);
  end]]
  --Blizzard made their tooltips update constantly.  Wasteful, but comparison tooltips will appear even if an edit box has focus.
  if GameTooltip:IsOwned(Button)then
    BaudManifestItemEntry_OnEnter(Button, true);
  end
end


function BaudManifestItemEntry_OnClick(Button)
  local Display = Button:GetParent();
  local ItemInfo = Display.Data[Button.Category][Button.ItemIndex];
  BaudManifestClearFocus();

  if ItemInfo.Category then
    if(arg1=="RightButton")then
      DebugMsg("Edit category name");
      BaudManifestCategoryEditBox:Hide();
      local Text = getglobal(Button:GetName().."Text");
      Text:Hide();
      --Unless the parent is set, the edit box can fall behind the frame
      BaudManifestCategoryEditBox:SetParent(Button);
      BaudManifestCategoryEditBox:SetFrameLevel(Button:GetFrameLevel() + 1);
      BaudManifestCategoryEditBox:SetPoint("LEFT",Text);
      BaudManifestCategoryEditBox:SetPoint("RIGHT",Text);
      BaudManifestCategoryEditBox.Display = Display;
      BaudManifestCategoryEditBox.Category = Display.Data[ItemInfo.Category];
      BaudManifestCategoryEditBox:SetText(Display.Data[ItemInfo.Category].Name);
      BaudManifestCategoryEditBox:Show();
      BaudManifestCategoryEditBox:SetFocus();
      BaudManifestCategoryEditBox:HighlightText();
    else
      DebugMsg("Edit category toggle expand");
      Display.Data[ItemInfo.Category].Expanded = not Display.Data[ItemInfo.Category].Expanded;
      BaudManifestUpdateDisplayList(Display);
    end
    return;
  end

  if(arg1=="LeftButton")then
    if IsControlKeyDown()then
      DressUpItemLink(ItemInfo.ItemString);
      return;
    end
    if IsShiftKeyDown()then
      if not ChatEdit_InsertLink(select(2,GetItemInfo(ItemInfo.ItemString..":"..(ItemInfo.SuffixFactor or 0))))then
        BaudManifestItemEntry_OnDragStart(Button);
      end
      return;
    end
    --Support for another addon
    if AltClickToAddItem and(strsub(this:GetName(),-8)=="BlizItem")and IsAltKeyDown()then
      ContainerFrameItemButton_OnModifiedClick(arg1);
    end
  end
end


function BaudManifestItemEntry_OnEnter(Button, Update)
  if not Update then
    BaudManifestItemHighlight:SetParent(Button);
    BaudManifestItemHighlight:SetFrameLevel(Button:GetFrameLevel() + 1);
    BaudManifestItemHighlight:SetAllPoints();
    BaudManifestItemHighlight:Show();
  end
  local Display = Button:GetParent();
  local ItemInfo = Display.Data[Button.Category][Button.ItemIndex];
  if not ItemInfo.ItemString then
    Button.updateTooltip = nil;
    return;
  end
  if not Update then
    --if this is updated repeatedly, it causes flickering
    BaudManifestTooltipIconTexture:SetTexture(select(10,GetItemInfo(ItemInfo.ItemString)));
    CooldownFrame_SetTimer(BaudManifestTooltipIconCooldown, GetItemCooldown(ItemInfo.ItemString));
    BaudManifestTooltipIcon:Show();

    BaudManifestItemHighlight:SetParent(Button);
    BaudManifestItemHighlight:SetFrameLevel(Button:GetFrameLevel() + 1);
  end
  if(Button:GetCenter() < UIParent:GetCenter())then
    GameTooltip:SetOwner(Button, "ANCHOR_RIGHT");
  else
    GameTooltip:SetOwner(Button, "ANCHOR_LEFT");
  end
  local SkipReset;
  local HasCooldown, RepairCost;
  if(Display:GetID() > 2)or(Display:GetID()==2)and not BankOpen or not ItemInfo.Slot then
    GameTooltip:SetHyperlink(ItemInfo.ItemString..":"..(ItemInfo.SuffixFactor or 0));
  elseif not ItemInfo.Bag then
    _, HasCooldown, RepairCost = GameTooltip:SetInventoryItem("player",ItemInfo.Slot);
  elseif(ItemInfo.Bag==-1)then
    _, HasCooldown, RepairCost = GameTooltip:SetInventoryItem("player",BankButtonIDToInvSlotID(ItemInfo.Slot));
  elseif(ItemInfo.Bag==-2)then
    _, HasCooldown, RepairCost = GameTooltip:SetInventoryItem("player",KeyRingButtonIDToInvSlotID(ItemInfo.Slot));
  else
    HasCooldown, RepairCost = GameTooltip:SetBagItem(ItemInfo.Bag,ItemInfo.Slot);
  end
  Button.updateTooltip = HasCooldown and TOOLTIP_UPDATE_TIME;

  if IsShiftKeyDown()then
    GameTooltip_ShowCompareItem();
  end

  if InRepairMode()and RepairCost and(RepairCost > 0)and IsDisplayAccessable(Display)then
    GameTooltip:AddLine(REPAIR_COST, "", 1, 1, 1);
    SetTooltipMoney(GameTooltip, RepairCost);
    GameTooltip:Show();
  elseif IsControlKeyDown()then
    ShowInspectCursor();
  elseif ItemInfo.Bag and ItemInfo.Slot and MerchantFrame:IsShown()and(MerchantFrame.selectedTab==1)and IsDisplayAccessable(Display)then
    ShowContainerSellCursor(ItemInfo.Bag,ItemInfo.Slot);
  else
    ResetCursor();
  end

  if ItemInfo.Equipped then
    GameTooltip:AddLine("Equipped");
    GameTooltip:Show();
  end

  if BaudManifestCategoriesDisabled(Display)and Display.Data[Button.Category].Path then
    GameTooltip:AddLine("Category: "..Display.Data[Button.Category].Path);
    GameTooltip:Show();
  end

  if(Display.Name=="All Characters")then
    local CharList = {};
    local Count;
    for Char, Value in pairs(RealmData)do
      if(Char~=Display.Name)then
        Count = 0;
        for Display, Value in ipairs(Value)do
          for Category, Value in ipairs(Value)do
            for Index, Value in ipairs(Value)do
              if(Value.ItemString==ItemInfo.ItemString)and Value.Count then
                Count = Count + Value.Count;
              end
            end
          end
        end
        if(Count > 0)then
          tinsert(CharList,{Name=Char,Count=Count});
        end
      end
    end
    if(#CharList > 0)then
      table.sort(CharList,function(a,b) return(a.Count > b.Count);end);
      for Key, Value in ipairs(CharList)do
        GameTooltip:AddLine(Value.Name..": "..Value.Count);
      end
      GameTooltip:Show();
    end
  end
end


function BaudManifestItemEntry_OnLeave()
  GameTooltip:Hide();
  BaudManifestItemHighlight:Hide();
  BaudManifestTooltipIcon:Hide();
  ResetCursor();
end


function  BaudManifestDisplayMoney_OnEnter()
  local Display = this:GetParent();
  if(Display.Name~="All Characters")then
    return;
  end
  local CharList = {};
  for Char, Value in pairs(RealmData)do
    if(Char~=Display.Name)then
      tinsert(CharList,{Name=Char, Gold=floor(Value.Money / 10000)})
    end
  end
  GameTooltip:SetOwner(this, "ANCHOR_TOPLEFT");
  GameTooltip:AddLine("Breakdown:",1,1,1);
  table.sort(CharList,function(a,b) return(a.Gold > b.Gold);end);
  for Key, Value in ipairs(CharList)do
    GameTooltip:AddLine(Value.Name..": "..Value.Gold.." gold");
  end
  GameTooltip:Show();
end


function  BaudManifestDisplayMoney_OnLeave()
  GameTooltip:Hide();
end


function BaudManifestSplitStackHandler(Button, Split)
  DragItem = {String = SplitItem, Origin = SplitOrigin, Remain = Split};
  BaudManifestPickupDragRemain();
end


--Quick split stops the addon from placing the item into an empty slot before moving it.  Best for selling and deleting.
function BaudManifestPickupDragRemain(MoveMethod, QuickSplit)
  if not DragItem or not DragItem.Remain then
    return;
  end
  if MoveMethod then
    if DragItem.MoveMethod and MoveMethod then
      DragItem = nil;
      ClearCursor();
      DebugMsg("PickupDragRemain:  Attempting to override existing move method.");
      return;
    end
    DragItem.MoveMethod = MoveMethod;
    DragItem.QuickSplit = QuickSplit;
  end
  local BestBag, BestSlot, BestCount, EmptyBag, EmptySlot;
  local Link, Count, Locked;
  local ItemString = DragItem.String;
  while true do
    if CursorHasItem()then
      DragItem = nil;
      ClearCursor();
      DebugMsg("PickupDragRemain:  Cursor has item already - aborting.");
      return;
    end
    if(DragItem.MoveMethod==true)and not BankOpen then
      DragItem = nil;
      ClearCursor();
      DebugMsg("PickupDragRemain:  Bank is not open - aborting.");
      return;
    end
    BestSlot, EmptySlot = nil, nil;
    BaudManifestForEachSlot(DragItem.Origin,function(Bag, Slot)
      if Bag then
        Link = GetContainerItemLink(Bag,Slot);
        _, Count, Locked = GetContainerItemInfo(Bag, Slot);
      else
        Link = GetInventoryItemLink("player",Slot);
        --Set count to zero so that items in your pack will be picked up before ones being worn, except when ctrl is being held
        Count = 1;--GetInventoryItemCount("player",Slot);
        Locked = IsInventoryItemLocked(Slot);
      end
      if Link then
        --This chooses stacks of the most appropriate size
        if not Locked and strfind(Link,ItemString..":",0,true)and(not BestSlot or(BestCount < DragItem.Remain)and(Count > BestCount)or(BestCount > DragItem.Remain)and(Count >= DragItem.Remain)and(Count < BestCount))then
          BestBag, BestSlot, BestCount = Bag, Slot, Count;
        end
      elseif Bag and not EmptySlot then
        --Empty slots are used for splitting items before moving them
        EmptyBag, EmptySlot = Bag, Slot;
      end
    end);
    if not BestSlot then
      DragItem = nil;
      DebugMsg("Could not find item for drag - aborting.");
      return;
    end
    --if the stack size isn't larger than what's requested, the whole thing is moved
    if(BestCount <= DragItem.Remain)then
      if not BestBag then
        IgnoreHooks = true;
        PickupInventoryItem(BestSlot);
        IgnoreHooks = false;
      elseif(DragItem.MoveMethod==true)then
        UseContainerItem(BestBag, BestSlot);  --Transfer to bank
      else
        IgnoreHooks = true;
        PickupContainerItem(BestBag, BestSlot);
        IgnoreHooks = false;
      end
      if(not BestBag or(DragItem.MoveMethod~=true))and not CursorHasItem()then
        DebugMsg("Item move failed - aborting.");
        DragItem = nil;
        return;
      end
      DragItem.Stack = BestCount;
      DragItem.Remain = DragItem.Remain - BestCount;
      DebugMsg("Picking up stack of "..BestCount.." from "..BestBag..","..BestSlot.." ("..DragItem.Remain.." remain)");
      if(DragItem.Remain <= 0)then
        DragItem.Remain = nil;
      end
    else
      DebugMsg("Splitting "..DragItem.Remain.." from stack at "..BestBag..","..BestSlot);
      IgnoreHooks = true;
      SplitContainerItem(BestBag, BestSlot, DragItem.Remain);
      IgnoreHooks = false;
      if not CursorHasItem()then
        DebugMsg("Failure to split stack.");
        DragItem = nil;
        return;
      end
      if not DragItem.QuickSplit and EmptySlot then
        --When a stack must be split, the moving procedure will continue after the move is finished
        DebugMsg("Placing split item at "..EmptyBag..","..EmptySlot);
        DragItem.Pending = true;
        IgnoreHooks = true;
        PickupContainerItem(EmptyBag, EmptySlot);
        IgnoreHooks = false;
        return;
      else
        --DragItem.Split = true;
        DragItem.Stack = DragItem.Remain;
        DragItem.Remain = nil;
        if(DragItem.MoveMethod==true)then
          BaudManifestPutItem(getglobal("BaudManifestDisplay"..(DragItem.Origin:GetID() % 2 + 1)));
        end
      end
    end
    --if MoveMethod is set to move between displays, no item will be in the cursor unless the item cannot be moved by using it
    if(not BestBag or(DragItem.MoveMethod~=true))and not CursorHasItem()then
      DebugMsg("Item move failed - aborting.");
      DragItem = nil;
      return;
    end
    if not DragItem.MoveMethod then
      BaudManifestShowDragCount();
      return;
    else
      if(DragItem.MoveMethod==true)and CursorHasItem()and not BaudManifestPutItem(getglobal("BaudManifestDisplay"..(DragItem.Origin:GetID() % 2 + 1)))or
          (DragItem.MoveMethod~=true)and not DragItem.MoveMethod()or not DragItem or not DragItem.Remain then
        ClearCursor();
        return;
      end
    end
  end
end


function BaudManifestItemEntry_OnDragStart(Button)
  local Display = Button:GetParent();
  if not BaudManifestDropItem(Display)then
    return;
  end
  local ItemInfo = Display.Data[Button.Category][Button.ItemIndex];
  if ItemInfo.Category then
    DebugMsg("Picking up category "..ItemInfo.Category);
    VirtualDrag = {Category = ItemInfo.Category, Origin = Display};
    BaudManifestVirtualDragIcon:SetTexture("Interface\\Icons\\INV_Scroll_03");
    BaudManifestVirtualDragIcon:SetAlpha(0.6);
    ClearCursor();  --Clear any spells and such in the cursor
    PlaySound("INTERFACESOUND_CURSORGRABOBJECT");
    BaudManifestVirtualDrag:Show();
    BaudManifestUpdateBlocked();
    return;
  end
  if ItemInfo.Count and(Display:GetID() <= 2)and((Display~=BaudManifestDisplay2)or BankOpen)then
    if(ItemInfo.Count > 1)and(IsShiftKeyDown()or IsAltKeyDown())then
      SplitItem = ItemInfo.ItemString;
      SplitOrigin = Display;
      if IsShiftKeyDown()and IsControlKeyDown()then
        BaudManifestSplitStackHandler(nil,ItemInfo.Count);
      elseif IsAltKeyDown()then
        BaudManifestSplitStackHandler(nil,1);
      else
        Button.SplitStack = BaudManifestSplitStackHandler;
        OpenStackSplitFrame(ItemInfo.Count, Button, "BOTTOMRIGHT", "TOPRIGHT");
      end
      return;
    end
    local ItemString = ItemInfo.ItemString;
    DebugMsg("Searching for "..ItemString);
    local Smallest = IsControlKeyDown();
    local BestBag, BestSlot, BestStack, Count, Locked, Link;
    BaudManifestForEachSlot(Display,
      function(Bag, Slot)
        if Bag then
          Link = GetContainerItemLink(Bag,Slot);
          _, Count, Locked = GetContainerItemInfo(Bag,Slot);
        else
          Link = GetInventoryItemLink("player",Slot);
          Count = GetInventoryItemCount("player",Slot);
          Locked = IsInventoryItemLocked(Slot);
        end
        if Link and strfind(Link,ItemString,0,true)and not Locked and(Bag or not BestBag)and(not BestBag or Smallest and
        (Count < BestStack)or not Smallest and(Count > BestStack))then
          BestBag, BestSlot, BestStack = Bag, Slot, Count;
        end
      end
    );
    if not BestStack then
      DebugMsg("Failed to find dragged item.");
    else
      if BestBag then
        --Drag item will automaticaly be set up since ignore hooks is not set
        PickupContainerItem(BestBag, BestSlot);
      else
        --Equipped items that are dragged from within the list are marked with the origin as the list so that when they are dropped,
        --they will be moved instead of placed in your inventory.  if they are dragged from the paper doll and dropped, they will
        --be unequipped.
        PickupInventoryItem(BestSlot);
        DragItem = {String = ItemString, Stack = GetInventoryItemCount("player",BestSlot), Origin = BaudManifestDisplay1};
        --BaudManifestShowDragCount();
      end
      if CursorHasItem()then
        DebugMsg("Picked up item.");
        return;
      else
        DebugMsg("Failed to pickup item.");
      end
    end
  end
  DebugMsg("Virtual item pickup");
  VirtualDrag = {ItemString = ItemInfo.ItemString, Origin = Display};
  BaudManifestVirtualDragIcon:SetTexture(select(10,GetItemInfo(ItemInfo.ItemString))or "Interface\\Icons\\INV_Misc_QuestionMark");
  BaudManifestVirtualDragIcon:SetAlpha(0.6);
  ClearCursor();  --Clear any spells and such in the cursor
  PlaySound("INTERFACESOUND_CURSORGRABOBJECT");
  BaudManifestVirtualDrag:Show();
  BaudManifestUpdateBlocked();
end


hooksecurefunc("SplitContainerItem",function(Bag, Slot, Amount)
  if IgnoreHooks then
    return;
  end
  SplitHoldItem = string.match(GetContainerItemLink(Bag, Slot)or "",ItemStringMatch);
  SplitHoldExpire = GetTime() + 120;
end);


function BaudManifestGetCursorPos(Display)
  local ScrollBar = getglobal(Display:GetName().."ScrollBar");
  local Position = (select(2,GetCursorPosition()) / (Display:GetEffectiveScale()) - ScrollBar:GetTop()) / -16 + 1;
  if(Position < 1)then
    Position = 1;
  end
  if(Position > Display.Entries + 1)then
    Position = Display.Entries + 1;
  end
  local DisplayIndex = floor(Position) + FauxScrollFrame_GetOffset(ScrollBar);
  if(DisplayIndex >= #Display.List + 1)then
    if(Position > #Display.List + 1)then
      Position = #Display.List + 1;
    end
    return Position, 0, 1, (#Display.Data[1] + 1);
  end
  local DisplayInfo = Display.List[DisplayIndex];
  local ItemIndex = DisplayInfo.Index;
  local ItemInfo = Display.Data[DisplayInfo.Category][ItemIndex];
  local Remain = Position % 1;
--[[  if ItemInfo.Category and(Remain > 0.25)and
  ((Remain < 0.75)or Display.List[DisplayIndex+1]and Display.List[DisplayIndex+1].Category==ItemInfo.Category)then
    return floor(Position + 1), DisplayInfo.Indent + 1, ItemInfo.Category, 1;]]
  if ItemInfo.Category and(Remain > 0.25)and
  ((Remain < 0.75)or Display.List[DisplayIndex+1]and Display.List[DisplayIndex+1].Category==ItemInfo.Category)then
    return floor(Position), DisplayInfo.Indent, ItemInfo.Category;

  elseif(Remain >= 0.5)then
    ItemIndex = ItemIndex + 1;
  end
  Position = floor(Position + 0.5);

  return Position, DisplayInfo.Indent, DisplayInfo.Category, ItemIndex;
end


function BaudManifestMoveListItem(Display)
  if(Display.Sort~=1)then
    UIErrorsFrame:AddMessage("List items cannot be arranged while the list is sorted.", 1.0, 0.1, 0.1, 1.0, UIERRORS_HOLD_TIME);
    return;
  end

  local _, _, NewCategory, NewIndex = BaudManifestGetCursorPos(Display);
  DebugMsg("Moving item to category "..NewCategory..", index "..(NewIndex or "-1"));

  --Recursive function for seeing if a category drop location is within itself
  local function FindCategory(Source)
    if(Source == NewCategory)then
      return true;
    end
    DebugMsg("Searching category #"..Source);
    for Key, Value in ipairs(Display.Data[Source])do
      if Value.Category and FindCategory(Value.Category)then
        return true;
      end
    end
  end

  if VirtualDrag and VirtualDrag.Category and FindCategory(VirtualDrag.Category)then
    UIErrorsFrame:AddMessage("You cannot place a category within itself.", 1.0, 0.1, 0.1, 1.0, UIERRORS_HOLD_TIME);
    return;
  end

  local OldItem;

  --Executed as a function so I can break out of both loops at once
  local function GetOldItem()
    for Category, Value in ipairs(Display.Data)do
      for ItemIndex, Value in ipairs(Value)do
        if Value.ItemString and(DragItem and(DragItem.String==Value.ItemString)or VirtualDrag and(VirtualDrag.ItemString==Value.ItemString))
        or Value.Category and VirtualDrag and(VirtualDrag.Category==Value.Category)then
          OldItem = Value;
          tremove(Display.Data[Category],ItemIndex);
          if NewIndex and(NewCategory==Category)and(NewIndex > ItemIndex)then
            NewIndex = NewIndex - 1;
          end
          return;
        end
      end
    end
  end
  GetOldItem();

  if not OldItem then
    DebugMsg("Failed to find move item");
    return;
  end
  if NewIndex then
    tinsert(Display.Data[NewCategory],NewIndex,OldItem);
  else
    tinsert(Display.Data[NewCategory],OldItem);
  end
  BaudManifestUpdateDisplayList(Display);
end


function BaudManifestPutItem(Display)
  local Type, ItemID = GetCursorInfo();
  if(Type~="item")then
    DebugMsg("Cursor is not carrying an item.");
    return;
  end
  local Result = true;
  local BestBag, BestSlot, BestPriority, Priority;
  BaudManifestForEachSlot(Display,
    function(Bag, Slot)
      if Bag then
        if(Slot==1)then
          Priority = BaudManifestBagPriority(Bag,ItemID);
        end
        if Priority and(not BestPriority or(Priority > BestPriority))and not GetContainerItemLink(Bag, Slot)then
          BestBag, BestSlot, BestPriority = Bag, Slot, Priority;
        end
      end
    end
  );
  if not BestBag then
    DebugMsg("Failed to find empty space.  Dropping item into bag to likely give an error.");
    if(Display:GetID()==1)then
      BestBag = 1;
    else
      BestBag = 5;
    end
    Result = nil;
  end

  DebugMsg("Dropping item at "..BestBag..", "..BestSlot..", priority "..BestPriority);
  if(BestBag==-1)then
    PutKeyInKeyRing();
  elseif(BestBag==0)then
    PickupContainerItem(BestBag, BestSlot);
  elseif(BestBag==1)then
    PutItemInBackpack();
  else
    PutItemInBag(ContainerIDToInventoryID(BestBag));
  end
  return Result;
end


function BaudManifestDropItem(Display)
  if not VirtualDrag and not CursorHasItem()then
    return true;
  end

  if VirtualDrag and(VirtualDrag.Origin~=Display)then
    UIErrorsFrame:AddMessage((VirtualDrag.Category and "Categories" or "That item").." cannot be moved between displays.", 1.0, 0.1, 0.1, 1.0, UIERRORS_HOLD_TIME);
    BaudManifestClearVirtualDrag();
    return;
  end
  if VirtualDrag or DragItem and(DragItem.Origin==Display)then
    --Dragging an item around in its own container will only move the item in the list
    BaudManifestMoveListItem(Display);
    BaudManifestClearVirtualDrag();
    ClearCursor();
    return;
  end
  if(Display:GetID() > 2)then
    UIErrorsFrame:AddMessage("Cannot put items into other char's manifests.", 1.0, 0.1, 0.1, 1.0, UIERRORS_HOLD_TIME);
    return;
  end
  if DragItem then
    ClearCursor();
    if not BankOpen then
      UIErrorsFrame:AddMessage("Cannot put items in the bank while the bank is closed.", 1.0, 0.1, 0.1, 1.0, UIERRORS_HOLD_TIME);
      return;
    end
    --Moving items between the bank and inventory is done by using them
    DragItem.Remain = (DragItem.Remain or 0) + DragItem.Stack;
    BaudManifestPickupDragRemain(true);
    return;
  end
  BaudManifestPutItem(Display);
end


function BaudManifestUpdateList(Display)
  if(Display:GetID() > 2)then
    --This should never happen, so it's a red flag if it does
    error("Attempt to update manifest of another character.");
    return;
  end
  if QueuedUpdate[Display:GetID()]then
    return;
  end
  if(Display:GetID()==2)and not BankOpen then
    DebugMsg("Not updating the bank box because it is not open.");
    return;
  end
  if SplitHoldItem and(GetTime() > SplitHoldExpire)then
    SplitHoldItem = nil;
  end
  DebugMsg("Updating item list "..Display:GetID()..".");

  --if the count is cleared here, an error during the update process could cause the count to not increase, and the items could be erased.
  for Category, Value in ipairs(Display.Data)do
    for ItemIndex, Value in ipairs(Value)do
      if Value.Slot then
        Value.Slot = nil;
      end
    end
  end
  local Incomplete = {};
  local Empty, Slots = 0, 0;
  local PickupBag, PickupSlot, PlaceBag, PlaceSlot;
  local ItemInfo, Link, ItemString, Suffix, UniqueID, Count, Quality, MaxCount, Locked, Found;
  local CountSlots;
  local CursorBusy = GetCursorInfo();
  local AddIndex = 1;

  BaudManifestForEachSlot(Display, function(Bag, Slot)
    if(Slot==1)then
      if not Bag or(Bag==-2)then
        CountSlots = false;
      elseif(Bag < 1)then
        CountSlots = true;
      else
        Link = GetInventoryItemLink("player",ContainerIDToInventoryID(Bag));
        CountSlots = Link and(select(7, GetItemInfo(Link))=="Bag");
      end
    end
    if CountSlots then
      Slots = Slots + 1;
    end
    if Bag then
      Link = GetContainerItemLink(Bag, Slot);
      _, Count, Locked = GetContainerItemInfo(Bag, Slot);
    else
      Link = GetInventoryItemLink("player", Slot);
      Count = GetInventoryItemCount("player", Slot);
      Locked = IsInventoryItemLocked(Slot);
    end
    if Link then
      ItemString, Suffix, UniqueID = string.match(Link,ItemStringMatch);
      if ItemString then
        Suffix = tonumber(Suffix);
        MaxCount = select(8,GetItemInfo(ItemString));
        ItemInfo = false;
        for Category, Value in ipairs(Display.Data)do
          for ItemIndex, Value in ipairs(Value)do
            if(Value.ItemString==ItemString)then
              if(not Value.Slot or not Value.Equipped and not Bag)then --and((Value.Equipped==true)~=(not Bag))then
                Value.Equipped = not Bag and true or nil;
              end
              if Value.Slot then
                Value.Count = Value.Count + Count;
                Value.Slots = (Value.Slots or 1) + 1;
              else
                Value.Count = Count;
                if Value.Slots then
                  Value.Slots = nil;
                end
              end

              if not Value.Slot or Bag and not Value.Bag or not Locked and Value.Locked or not Locked and(Count > Value.Stack)then
                Value.Bag = Bag;
                Value.Slot = Slot;
                Value.Stack = Count;
                Value.Locked = Locked;
              end
              ItemInfo = Value;
              break;
            end
          end
        end
        if not ItemInfo then
          ItemInfo = {ItemString=ItemString,Count=Count,Bag=Bag,Slot=Slot,Locked=Locked,Stack=Count};
          if not Bag then
            ItemInfo.Equipped = true;
          end
          Quality = select(3, GetItemInfo(ItemString));
          tinsert(Display.Data[(Display:GetID()==1)and Quality and PlayerData.PutQuality[Quality]or 1],ItemInfo);
        end

        --When the item suffix is negative, that means there is a scaling factor stored in the first 16 bits of the UniqueID.
        --Without this scaling factor, the item's stats will not display correctly if a tooltip is created from the info.
        --As of patch 2.1, unless the remaining bits of the UniqueID are kept, linked items will disconnect the player.
        if ItemInfo.SuffixFactor or(Suffix < 0)then
          ItemInfo.SuffixFactor = (Suffix < 0)and UniqueID or nil; --bit.band(tonumber(UniqueID),65535);
        end

        --MaxCount can register nil here
        if Bag and not CursorBusy and(PlayerData.Combine~=false)and MaxCount and(Count < MaxCount)and not Locked and(SplitHoldItem~=ItemString)and(not DragItem or(DragItem.String~=ItemString)or not DragItem.Pending)then
          if not Incomplete[ItemString]then
            Incomplete[ItemString] = {Bag=Bag, Slot=Slot, Count=Count};
          else
            PickupBag, PickupSlot = Bag, Slot;
            PlaceBag, PlaceSlot = Incomplete[ItemString].Bag,Incomplete[ItemString].Slot;
            --Combine two incomplete stacks by moving the larger one onto the smaller
            if(Count < Incomplete[ItemString].Count)then
              PickupBag, PickupSlot, PlaceBag, PlaceSlot = PlaceBag, PlaceSlot, PickupBag, PickupSlot
            end
            Incomplete[ItemString] = nil;
            DebugMsg("Item is "..(Locked and "locked" or "unlocked"));
            ClearCursor();
            DebugMsg("Combining incomplete stacks.");
            IgnoreHooks = true;
            PickupContainerItem(PickupBag, PickupSlot);
            if CursorHasItem()then
              PickupContainerItem(PlaceBag, PlaceSlot);
              ClearCursor();
            end
            IgnoreHooks = false;
          end
        end
      end
    elseif CountSlots then
      Empty = Empty + 1;
    end
  end);
  --Set counts of items that were not found to 0.  Safest to be done after, in case of errors occuring, causing the scan process to end prematurely.
  for Category, Value in ipairs(Display.Data)do
    for ItemIndex, Value in ipairs(Value)do
      if Value.Count and not Value.Slot then
        Value.Count = nil;
        Value.Equipped = nil;
        Value.Slots = nil;
        Value.Show = true;
        Display.ClearZero = true;
      end
    end
  end
  Display.Data.SlotEmpty = Empty;
  Display.Data.SlotTotal = Slots;
  BaudManifestUpdateDisplayList(Display);
end


function BaudManifestCountCategorySlots(Display)
  DebugMsg("Counting slot usage for all categories.");
  for Key, Value in ipairs(Display.Data)do
    Value.SlotsUsed = nil;
  end

  local function CountCategory(Category, Path)
    Category.SlotsUsed = 0;
    Category.Path = Path;
    local Index = 1;
    while Category[Index]do
      if Category[Index].ItemString and Category[Index].Count then
        Category.SlotsUsed = Category.SlotsUsed + (Category[Index].Slots or 1);
      elseif Category[Index].Category then
        if not Display.Data[Category[Index].Category].SlotsUsed then
          Category.SlotsUsed = Category.SlotsUsed + CountCategory(Display.Data[Category[Index].Category], (Path and Path.."\\" or "")..Display.Data[Category[Index].Category].Name);
        else
          DEFAULT_CHAT_FRAME:AddMessage("BaudManifest: Removing duplicate category entry.",1,1,0);
          tremove(Category,Index);
          Index = Index - 1;
        end
      end
      Index = Index + 1;
    end
    return Category.SlotsUsed;
  end
  CountCategory(Display.Data[1]);

  local Found;
  for Category, Value in ipairs(Display.Data)do
    if not Value.SlotsUsed then
      DEFAULT_CHAT_FRAME:AddMessage("BaudManifest: Found stray category - restoring.",1,1,0);
      Found = true;
      tinsert(Display.Data[1],{Category = Category});
    end
  end
  if Found then
    return BaudManifestCountCategorySlots(Display);
  end
end


function BaudManifestCustomFilter(ItemString, Filter)
  if Filter and strfind(Filter,"[^%s]")then
    Filter = strlower(Filter);
    BaudManifestTooltip:SetHyperlink(ItemString);
    for Index = 1, BaudManifestTooltip:NumLines()do
--      ChatFrame1:AddMessage(getglobal("BaudManifestTooltipTextLeft"..Index):GetText()or "nil");
      if strfind(strlower(getglobal("BaudManifestTooltipTextLeft"..Index):GetText()or ""),Filter,1,true)then
        return true;
      end
      if strfind(strlower(getglobal("BaudManifestTooltipTextRight"..Index):GetText()or ""),Filter,1,true)then
        return true;
      end
    end
    return false;
  end
  return true;
end


function BaudManifestUpdateDisplayList(Display)
  if not Display:IsShown()then
    DebugMsg("Not updating display list while the list is closed.");
    return;
  end

  DebugMsg("Generating display list.");
  BaudManifestCountCategorySlots(Display);
  local Sort;
  Display.List = {};
  Display.UpdateList = nil;

  local HideCategories = BaudManifestCategoriesDisabled(Display);

  local function AddCategory(Category, Indent)
    for Key, Value in ipairs(Display.Data[Category])do
      if Value.Category then
        if not HideCategories then
          tinsert(Display.List,{Index = Key, Category = Category, Indent = Indent});
          if Display.Data[Value.Category].Expanded then
            AddCategory(Value.Category, Indent + 1);
          end
        else
          AddCategory(Value.Category,0);
        end
      elseif Value.ItemString and(Value.Count or Value.Show)and((PlayerData.HideEquipped==false)or HideCategories or(Display:GetID() > 2)or not Value.Equipped or(Value.Count > 1))and
      (BaudManifestCustomFilter(Value.ItemString, Display.CustomFilter)and((Display.Filter==1)or(Display.Filter==select(6,GetItemInfo(Value.ItemString)))))then
        if(Display.Sort==1)then
          tinsert(Display.List,{Index = Key, Category = Category, Indent = Indent});
        else
          --Items that enter your backpack which the client has never seen before will not have a lot of GetItemInfo stuff, so
          --unsortable entries will not be shown
          Sort = Sorts[Display.Sort].Func(Value);
          if Sort then
            tinsert(Display.List,{Index = Key, Category = Category, Indent = Indent, Sort = Sort});
          end
        end
      end
    end
  end
  AddCategory(1,0);

  if(Display.Sort~=1)then
    table.sort(Display.List,function(a,b) return(a.Sort==b.Sort)and(a.Index < b.Index)or(a.Sort < b.Sort); end);
  end
  DebugMsg("Display list finished.  Updating visual.");
  BaudManifestScrollBar_Update(Display);
end


function BaudManifestScrollBar_Update(Display)
  if not Display then
    Display = this:GetParent();
  end
  if not Display:IsShown()or not Display.List then
    return;
  end
  if not Display.UpdateList then
    Display.Refresh = nil;
  end
  if(BaudManifestCategoryEditBox.Display==Display)then
    BaudManifestCategoryEditBox:Hide();
  end
  local Index, Name, Quality, MaxCount, Texture, R, G, B, Text;
  local Button, BlizItem, BlizInv, Cooldown, ButtonText, Icon;

  Text = getglobal(Display:GetName().."Text");
  Name = DisplayNames[(Display:GetID()-1)%2 + 1];
  if Display.Data.SlotEmpty then
    Name = Name.." ("..Display.Data.SlotEmpty.."/"..Display.Data.SlotTotal..")";
  end
  if(Display:GetID()==1)or(Display:GetID()==2)and BankOpen then
    Text:SetTextColor(NORMAL_FONT_COLOR.r,NORMAL_FONT_COLOR.g,NORMAL_FONT_COLOR.b);
  else
    Name = Name.." - Offline";
    Text:SetTextColor(RED_FONT_COLOR.r,RED_FONT_COLOR.g,RED_FONT_COLOR.b);
  end
  Text:SetText(Name);
  local FrameName = Display:GetName();
  local ScrollBar = getglobal(FrameName.."ScrollBar");
  local Total = #Display.List;
  FauxScrollFrame_Update(ScrollBar,Total,Display.Entries,16);
  local NoEntries = getglobal(Display:GetName().."NoEntries");
  if(Total<=0)then
    NoEntries:Show();
  else
    NoEntries:Hide();
  end
  local ItemInfo;
  for Line = 1, Display.Entries do
    Index = Line + FauxScrollFrame_GetOffset(ScrollBar);
    Button = getglobal(FrameName.."Entry"..Line);
    if(Index > Total)then
      Button:Hide();
    else
      Button.ItemIndex = Display.List[Index].Index;
      Button.Category = Display.List[Index].Category;
      ItemInfo = Display.Data[Button.Category][Button.ItemIndex];
      BlizItem = getglobal(Button:GetName().."BlizItem");
      BlizInv = getglobal(Button:GetName().."BlizInv");
      Icon = getglobal(Button:GetName().."Texture");
      Cooldown = getglobal(Button:GetName().."Cooldown");
      ButtonText = getglobal(Button:GetName().."Text");
      ButtonText:Show();
      Button:SetPoint("LEFT",ScrollBar,"LEFT",Display.List[Index].Indent * IndentSize,0);
      if ItemInfo.ItemString then
        Button:SetNormalTexture("");
        Button:SetPushedTexture("");
        Button:SetHighlightTexture("");
        Name, _, Quality, _, _, _, _, MaxCount, _, Texture = GetItemInfo(ItemInfo.ItemString);
        if not Name then
          Name, Quality, MaxCount, Texture = "Unknown", 0, 1, "Interface\\Icons\\INV_Misc_QuestionMark";
          Display.Refresh = GetTime() + 1;
        end
        CooldownFrame_SetTimer(Cooldown, GetItemCooldown(ItemInfo.ItemString));
        if not ItemInfo.Slot then
          BlizItem:Hide();
          BlizInv:Hide();
        elseif ItemInfo.Bag then
          BlizInv:Hide();
          Button:SetID(ItemInfo.Bag);
          BlizItem:SetID(ItemInfo.Slot);
          BlizItem:Show();
        else
          BlizItem:Hide();
          BlizInv:SetID(ItemInfo.Slot);
          BlizInv:Show();
        end
        if(MaxCount > 1)or(ItemInfo.Count~=1)then
          Text = ItemInfo.Count or 0;
          if ItemInfo.Slots then
            Text = Text.."("..ItemInfo.Slots..")";
          end
          if PlayerData.ItemNameFirst then
            Name = Name.." "..Text;
          else
            Name = Text.." "..Name;
          end
        end
        if ItemInfo.Equipped then
          Name = "|cffffD100E|r "..Name;
        end
        ButtonText:SetText(Name);
        if ItemInfo.Count then
          R, G, B = GetItemQualityColor(Quality or 1);
          ButtonText:SetTextColor(R, G, B, 1);
        else
          ButtonText:SetTextColor(0.2,0.2,0.2,1);
        end
        if ItemInfo.Count and(not ItemInfo.Locked or(Display:GetID() > 2))and((Display:GetID()~=2)or BankOpen)then
          Icon:SetVertexColor(1,1,1);
        else
          Icon:SetVertexColor(0.5,0.5,0.5);
        end
        Icon:SetTexture(Texture);
        Icon:Show();
        if GameTooltip:IsOwned(Button)then
          BaudManifestItemEntry_OnEnter(Button);
        end
        Button:Show();
      else
        BlizItem:Hide();
        BlizInv:Hide();
        ButtonText:SetText(Display.Data[ItemInfo.Category].Name.." ("..Display.Data[ItemInfo.Category].SlotsUsed..")");
        ButtonText:SetTextColor(NORMAL_FONT_COLOR.r,NORMAL_FONT_COLOR.g,NORMAL_FONT_COLOR.b,1);
        Icon:Hide();
        Cooldown:Hide();
        Text = Display.Data[ItemInfo.Category].Expanded and "Minus" or "Plus";
        Button:SetNormalTexture("Interface\\Buttons\\UI-"..Text.."Button-Up");
        Button:SetPushedTexture("Interface\\Buttons\\UI-"..Text.."Button-Down");
        Button:SetHighlightTexture("Interface\\Buttons\\UI-PlusButton-Hilight");
        Button:Show();
      end
    end
  end
end


function BaudManifestOptionsButton_OnClick()
  ToggleDropDownMenu(1, this:GetParent(), BaudManifestOptionsDropDown, this:GetName(), 0, 0);
end


function BaudManifestOptionsDropDown_Initialize()
  local Display = UIDROPDOWNMENU_MENU_VALUE;
  if not Display then
    return;
  end
  local info = {}; --UIDropDownMenu_CreateInfo();
  info.arg1 = Display;

  if(Display:GetID()==1)and not BaudManifestDisplay2:IsShown()then
    info.text = "Show Bank";
    info.func = function()
      BaudManifestDisplay2:Show();
    end
    UIDropDownMenu_AddButton(info);
  end

  info.text = "Create Category";
  info.func = BaudManifestGetCategory;
  UIDropDownMenu_AddButton(info);

  info.text = "Categorize By Type";
  info.func = function(Display)
    if BaudManifestCategoriesDisabled(Display)then
      UIErrorsFrame:AddMessage("Cannot create categories while the list is sorted or filtered.", 1.0, 0.1, 0.1, 1.0, UIERRORS_HOLD_TIME);
      return;
    end
    local Dialog = StaticPopup_Show("BAUDMANIFEST_CATEGORIZE");
    if Dialog then
      Dialog.data = Display;
    end
  end
  UIDropDownMenu_AddButton(info);

  if(Display:GetID() <= 2)then
    info.text = "Lock Display Size";
    info.checked = (Display.Data.LockSize and 1 or nil);
    info.func = function(Display)
      Display.Data.LockSize = not Display.Data.LockSize;
    end
    UIDropDownMenu_AddButton(info);
  end

  info.text = "Background Color";
  info.checked = nil;
  info.hasColorSwatch = 1;
  info.r, info.g, info.b, info.opacity = unpack(Display.Data.BGColor);
  info.opacity = 1 - info.opacity;
  info.swatchFunc = function()
    local BGColor = ColorPickerFrame.arg1.Data.BGColor;
    BGColor[1], BGColor[2], BGColor[3] = ColorPickerFrame:GetColorRGB();
    BaudManifestUpdateBGColor(ColorPickerFrame.arg1);
  end
  info.func = function()
    UIDropDownMenuButton_OpenColorPicker();
  end
  info.hasOpacity = 1;
  info.opacityFunc = function()
    ColorPickerFrame.arg1.Data.BGColor[4] = 1.0 - OpacitySliderFrame:GetValue();
    BaudManifestUpdateBGColor(ColorPickerFrame.arg1);
  end
  info.cancelFunc = function(previousValues)
    ColorPickerFrame.arg1.Data.BGColor = {previousValues.r, previousValues.g, previousValues.b, 1 - previousValues.opacity};
    BaudManifestUpdateBGColor(ColorPickerFrame.arg1);
  end
  UIDropDownMenu_AddButton(info);

  local info = {}; --UIDropDownMenu_CreateInfo();
  info.text = "Other Characters...";
  info.func = function()
    BaudManifestCharacters:Show();
  end
  UIDropDownMenu_AddButton(info);

  info.text = "Options...";
  info.func = function()
    BaudManifestOptions:Show();
  end
  UIDropDownMenu_AddButton(info);
end


--This hook function is needed to pass variables to the color picker (to determine the display)
--Information must be passed before the function finishes executing, which prevents me from using a secure handler
local pre_UIDropDownMenuButton_OpenColorPicker = UIDropDownMenuButton_OpenColorPicker;
UIDropDownMenuButton_OpenColorPicker = function(Button)
  if not Button then
    Button = this;
  end
  ColorPickerFrame.arg1 = Button.arg1;
  pre_UIDropDownMenuButton_OpenColorPicker(Button);
end


function BaudManifestUpdateBGColor(Display)
  Display:SetBackdropColor(unpack(Display.Data.BGColor));
end


function BaudManifestSortDropDown_Initialize()
  --Note that when this function runs when the addon is loaded, it determines the text to be displayed on the drop down
  local Display = getglobal(UIDROPDOWNMENU_INIT_MENU):GetParent();

  local info = {}; --UIDropDownMenu_CreateInfo();
  info.func = BaudManifestSortDropDown_OnClick;

  for Key, Value in ipairs(Sorts)do
    info.text = Value.Name;
    info.value = Key;
    info.checked = (info.value == Display.Sort)and 1 or nil;
    UIDropDownMenu_AddButton(info);
  end

  local Text = getglobal(UIDROPDOWNMENU_INIT_MENU.."Text");
end


function BaudManifestSortDropDown_OnClick()
  local Display = getglobal(UIDROPDOWNMENU_OPEN_MENU):GetParent();
  Display.Sort = this.value;
  UIDropDownMenu_SetSelectedValue(getglobal(UIDROPDOWNMENU_OPEN_MENU), this.value);
  local Text = getglobal(UIDROPDOWNMENU_OPEN_MENU.."Text");
  if(this.value==1)then
    Text:SetTextColor(1,1,1);
  else
    Text:SetTextColor(NORMAL_FONT_COLOR.r,NORMAL_FONT_COLOR.g,NORMAL_FONT_COLOR.b);
  end
  BaudManifestUpdateDisplayList(Display);
end


function BaudManifestFilterDropDown_Initialize()
  --Note that when this function runs when the addon is loaded, it determines the text to be displayed on the drop down
  local Display = getglobal(UIDROPDOWNMENU_INIT_MENU):GetParent();

--  SetTextColor();
  local info = {}; --UIDropDownMenu_CreateInfo();
  info.func = BaudManifestFilterDropDown_OnClick;

  info.text = "No Filter";
  info.value = 1;
  info.checked = (info.value == Display.Filter)and 1 or nil;
  UIDropDownMenu_AddButton(info);

  if not Display.Data then
    return;
  end

  local Text = getglobal(UIDROPDOWNMENU_INIT_MENU.."Text");
  if(Display.Filter~=1)then
    Text:SetTextColor(NORMAL_FONT_COLOR.r,NORMAL_FONT_COLOR.g,NORMAL_FONT_COLOR.b);
  else
    Text:SetTextColor(1,1,1);
  end

  local ItemTypes = {};
  local Type;
  for Category, Value in ipairs(Display.Data)do
    for ItemIndex, Value in ipairs(Value)do
      if Value.ItemString then
        Type = select(6,GetItemInfo(Value.ItemString));
        for Key, Value in ipairs(ItemTypes)do
          if(Type==Value)then
            Type = nil;
            break;
          end
        end
        if Type then
          tinsert(ItemTypes,Type);
        end
      end
    end
  end

  table.sort(ItemTypes,function(a,b) return (a < b);end);

  for Key, Value in ipairs(ItemTypes)do
    info.text = Value;
    info.value = Value;
    info.checked = (info.value == Display.Filter)and 1 or nil;
    UIDropDownMenu_AddButton(info);
  end
end


function BaudManifestFilterDropDown_OnClick()
  local Display = getglobal(UIDROPDOWNMENU_OPEN_MENU):GetParent();
  Display.Filter = this.value;
  UIDropDownMenu_SetSelectedValue(getglobal(UIDROPDOWNMENU_OPEN_MENU), this.value);
  local Text = getglobal(UIDROPDOWNMENU_OPEN_MENU.."Text");
  if(this.value==1)then
    Text:SetTextColor(1,1,1);
  else
    Text:SetTextColor(NORMAL_FONT_COLOR.r,NORMAL_FONT_COLOR.g,NORMAL_FONT_COLOR.b);
  end
  BaudManifestUpdateDisplayList(Display);
end


function BaudManifestFilterEditBox_OnTextChanged()
  local Display = this:GetParent();
  if((Display.CustomFilter or "")==this:GetText())then
    return;
  end
  Display.CustomFilter = this:GetText();
  Display.Refresh = GetTime() + FilterDelay;
  Display.UpdateList = true;
end


function BaudManifestCharacters_OnLoad()
  this:RegisterForClicks("RightButtonUp");
  tinsert(UIMenus,this:GetName());
end


function BaudManifestCharacters_OnClick()
  BaudManifestClearVirtualDrag();
end


function BaudManifestCharacters_OnShow()
--  BaudManifestCharactersScrollBoxHighlight:Hide();
end


function BaudManifestCharacterEntry_OnClick()
  BaudManifestClearVirtualDrag();
  SelectedChar = this.Name;
  if(SelectedChar=="All Characters")then
    BaudManifestUpdateAllChars();
  end
  BaudManifestCharactersScrollBoxHighlight:SetPoint("TOP",this);
  BaudManifestCharactersScrollBoxHighlight:Show();
  local Display;
  for Index = 1, 2 do
    Display = getglobal("BaudManifestDisplay"..(Index + 2));
    if(Index~=1)and(SelectedChar=="All Characters")then
      Display:Hide();
      break;
    end
    getglobal(Display:GetName().."CloseButton"):Hide();
    getglobal(Display:GetName().."OptionsButton"):SetPoint("TOPRIGHT",-5,-5);
    Display.Name = SelectedChar;
    Display.Data = RealmData[SelectedChar][Index];
    BaudManifestClearZeros(Display);
    BaudManifestUpdateBGColor(Display);
    BaudManifestUpdateDisplayList(Display);
    if(Index==1)then
      SetMoneyFrameColor(Display:GetName().."MoneyFrame", 1.0, 1.0, 1.0);
      MoneyFrame_Update(Display:GetName().."MoneyFrame", (RealmData[SelectedChar].Money or 0));
    end
    Display:Show();
  end
end


function BaudManifestUpdateAllChars()
  DebugMsg("Updating combined character list");
  local AllChars = GetCharData("All Characters");
  AllChars.Money = 0;

  for Category, Value in ipairs(AllChars[1])do
    for ItemIndex, Value in ipairs(Value)do
      Value.Count = 0;
      Value.Slots = nil;
    end
  end

  local function AddItem(Item)
    for Category, Value in ipairs(AllChars[1])do
      for Key, Value in ipairs(Value)do
        if(Value.ItemString==Item.ItemString)then
          Value.Count = (Value.Count or 0) + Item.Count;
          return true;
        end
      end
    end
  end

  for Key, Value in pairs(RealmData)do
    if(Value~=AllChars)then
      AllChars.Money = AllChars.Money + (Value.Money or 0);
      for Display, Value in ipairs(Value)do
        for Category, Value in ipairs(Value)do
          for Key, Value in ipairs(Value)do
            if Value.ItemString and Value.Count and(Value.Count > 0)and not AddItem(Value)then
              tinsert(AllChars[1][1],{ItemString = Value.ItemString, Count = Value.Count, SuffixFactor = Value.SuffixFactor});
            end
          end
        end
      end
    end
  end

  for Category, Value in ipairs(AllChars[1])do
    for ItemIndex, Value in ipairs(Value)do
      if(Value.Count==0)then
        Value.Count = nil;
        Value.Show = true;
      end
    end
  end
end


function BaudManifestOptions_OnLoad()
  tinsert(UIMenus,this:GetName());
  local Text, DropDown;
  local QText = {[0] = "grey", [1] = "white", [2] = "green"}
  for Quality = 0, 2 do
    DropDown = CreateFrame("Frame",this:GetName().."QualityDropDown"..Quality,this,"UIDropDownMenuTemplate");
    DropDown:SetPoint("TOPLEFT",155,-35-Quality*30);
    DropDown:SetID(Quality);
    UIDropDownMenu_SetWidth(110, DropDown);
    Text = this:CreateFontString(this:GetName().."QualityText"..Quality,"ARTWORK","GameFontHighlight");
    Text:SetText("Put new "..select(4,GetItemQualityColor(Quality))..QText[Quality].."|r items in");
    Text:SetPoint("RIGHT",DropDown,"LEFT",10,0);
  end
  local Button;
  for Index, Value in ipairs(CheckButtons)do
    Button = CreateFrame("CheckButton",this:GetName().."CheckButton"..Index,this,"OptionsCheckButtonTemplate");
    Button:SetPoint("TOPRIGHT",this,-160+(Value.Depend and 10 or 0),-25*Index);
    Button:SetID(Index);
    Button.tooltipText = Value.TooltipText;
    getglobal(Button:GetName().."Text"):SetText(Value.Text);
    Button:SetScript("OnClick",BaudManifestCheckButton_OnClick);
  end
end


local BankBlocked = false;
function BaudManifestUpdateBlizBankBlock()
  local Block = PlayerData.ShowBank and PlayerData.HideBlizBank;
  if(BankBlocked==Block)then
    return;
  end
  DebugMsg("Bliz bank blocked: "..(Block and "On" or "Off"));
  BankBlocked = Block;
  if Block then
    BankFrame:UnregisterEvent("BANKFRAME_OPENED");
  else
    BankFrame:RegisterEvent("BANKFRAME_OPENED");
  end
end


function BaudManifestOptions_OnShow()
  local DropDown;
  for Quality = 0, 2 do
    DropDown = getglobal("BaudManifestOptionsQualityDropDown"..Quality);
    UIDropDownMenu_Initialize(DropDown, BaudManifestQualityDropDown_Initialize);
    UIDropDownMenu_SetSelectedValue(DropDown, PlayerData.PutQuality[Quality]);
  end
  local Button;
  for Index, Value in ipairs(CheckButtons)do
    Button = getglobal(this:GetName().."CheckButton"..Index);
    Button:SetChecked(PlayerData[Value.SavedVar]);
  end
  BaudManifestOptionsUpdateDepend();

  BaudManifestOptionsScaleSliderLow:SetText("50%");
  BaudManifestOptionsScaleSliderHigh:SetText("200%");
  BaudManifestOptionsScaleSlider.tooltipText = "Scale of the manifest.";
  BaudManifestOptionsScaleSlider:SetValue(PlayerData.Scale);
end


function BaudManifestOptionsUpdateDepend()
  for Index, Value in ipairs(CheckButtons)do
    if Value.Depend then
      Button = getglobal("BaudManifestOptionsCheckButton"..Index);
      if PlayerData[Value.Depend]then
        OptionsFrame_EnableCheckBox(Button);
      else
        OptionsFrame_DisableCheckBox(Button);
      end
    end
  end
end


function BaudManifestQualityDropDown_Initialize()
  local Quality = getglobal(UIDROPDOWNMENU_INIT_MENU):GetID();

  local info = {}; --UIDropDownMenu_CreateInfo();
  info.func = BaudManifestQualityDropDown_OnClick;

  local CategoryList = {};
  for Category, Value in ipairs(PlayerData[1])do
    if(Category > 1)then
      tinsert(CategoryList,{Name=Value.Name,Category=Category});
    end
  end
  table.sort(CategoryList,function(a,b)return(a.Name > b.Name)end);
  tinsert(CategoryList,1,{Name="<Inventory>",Category=1});

  for Index, Value in ipairs(CategoryList)do
    info.text = Value.Name;
    info.value = Value.Category;
    info.checked = (info.value == PlayerData.PutQuality[Quality])and 1 or nil;
    UIDropDownMenu_AddButton(info);
  end
end


function BaudManifestQualityDropDown_OnClick()
  PlayerData.PutQuality[getglobal(UIDROPDOWNMENU_OPEN_MENU):GetID()] = this.value;
  UIDropDownMenu_SetSelectedValue(getglobal(UIDROPDOWNMENU_OPEN_MENU), this.value);
end


function BaudManifestCheckButton_OnClick()
  if this:GetChecked()then
    PlaySound("igMainMenuOptionCheckBoxOff");
  else
    PlaySound("igMainMenuOptionCheckBoxOn");
  end
  local SavedVar = CheckButtons[this:GetID()].SavedVar;
  PlayerData[SavedVar] = this:GetChecked()and true or false;
  if(SavedVar=="HideEquipped")then
    BaudManifestUpdateDisplayList(BaudManifestDisplay1);
  elseif(SavedVar=="ItemNameFirst")then
    for Display = 1, 2 do
      BaudManifestScrollBar_Update(getglobal("BaudManifestDisplay"..Display));
    end
  end
  BaudManifestUpdateBlizBankBlock();
  BaudManifestOptionsUpdateDepend();
end


function BaudManifestUpdateScales()
  local Display, X, Y, Multiplier;
  for Index = 1, 4 do
    Display = getglobal("BaudManifestDisplay"..Index);
    Multiplier = Display:GetScale() / (PlayerData.Scale / 100);
    X, Y = Display:GetCenter();
    Display:SetScale(PlayerData.Scale / 100);
    if(Index <= 2)then
      Display:ClearAllPoints();
      Display:SetPoint("CENTER",UIParent,"BOTTOMLEFT", X * Multiplier, Y * Multiplier);
    end
    --Fix for the fact that the height registers as 0 after scaling
    Display.UpdateHeight = true;
    --BaudManifestDisplay_OnSizeChanged(Display);
  end
end


function BaudManifestScaleSlider_OnValueChanged()
  getglobal(this:GetName().."Text"):SetText("Scale: "..this:GetValue().."%");
  if IgnoreHooks then
    return;
  end
  PlayerData.Scale = this:GetValue();
  BaudManifestUpdateScales();
end


local pre_OpenBackpack = OpenBackpack;
OpenBackpack = function()
  if not PlayerData.HideBlizInv and((this==MailFrame)or(this==MerchantFrame))then
    pre_OpenBackpack();
  end
end